<?php
// phpXMLDOM - an XML DOM Implementation
//
// This Software and all associated files are released unter the 
// GNU Public License (GPL), see LICENSE for details.
//
// $Id: Document.php,v 1.6 2002/02/13 22:47:40 thomi Exp $

/**
 * Class representing the DOM Document interface.
 *
 * @package  phpXD
 * @author   Thomas Dohmke <thomas@dohmke.de>
 * @version  $Revision: 1.6 $
 */
class Document extends Node {
  /**
   * An instance of the DocumentType interface representing the DTD; If no
   * DTD was declared in the document, this property is null.<br>
   * DOM-Level 1
   *
   * @private
   * @type    DocumentType
   */
  var $doctype;
  
  /**
   * Points to the single Element node that is the root of the XML tree.<br>
   * DOM-Level 1
   *
   * @private
   * @type    Element
   */
  var $documentElement;
  
  /**
   * An reference to the DOMImplementation that is responsible for this 
   * document.<br>
   * DOM-Level 1
   *
   * @private
   * @type    DOMImplementation
   */
  var $implementation;
  
  /**
   * Constructor of the class.<br>
   *
   * @public
   */
  function Document() {
    $this->Node();
    $this->nodeType = DOCUMENT_NODE;
    $this->nodeName = "#document";
    $this->implementation = new DOMImplementation();
    $this->ownerDocument =& $this;
  }
  
  /**
   * Returns an reference to the DocumentType object.<br>
   * DOM-Level 1
   *
   * @public
   * @returns DocumentType
   */
  function &getDoctype() {
    return $this->doctype;
  }
  
  /**
   * Returns an reference to the root element.<br>
   * DOM-Level 1
   *
   * @public
   * @returns Element
   */
  function &getDocumentElement() {
    return $this->documentElement;
  }
  
  /**
   * Returns an reference to the DOMImplementation object.<br>
   * DOM-Level 1
   *
   * @public
   * @returns DOMImplementation
   */
  function &getImplementation() {
    return $this->implementation;
  }
  
  /**
   * Creates an Attribute node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name  <code>string</code>
   * @param   $value <code>string</code>
   * @returns Attr
   */
  function &createAttribute($name) {
    $new = new Attr();
    $new->nodeName = $name;
    $new->ownerElement = NULL;
    $new->ownerDocument =& $this;
    return $new;
  }
  
  /**
   * Creates a CDataSection node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $data 
   * @returns CDataSection
   */
  function &createCDataSection($data) {
    $new = new CDataSection();
    $new->nodeValue = $data;
    $new->ownerDocument =& $this;
    return $new;
  }
  
  /**
   * Creates a Comment node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $data <code>string</code>
   * @returns Comment
   */
  function &createComment($data) {
    $new = new Comment();
    $new->nodeValue = $data;
    $new->ownerDocument =& $this;
    return $new;
  }
  
  /**
   * Creates a DocumentFragment node.<br>
   * DOM-Level 1
   *
   * @public
   * @returns DocumentFragment
   */
  function &createDocumentFragment() {
    $new = new DocumentFragment();
    return $new;
  }
  
  /**
   * Creates an Element node.<br>
   * DOM-Level 1
   *
   * @public
   * @param  $tagName <code>string</code>
   * @returns Element
   */
  function &createElement($tagName) {
    $new = new Element();
    $new->nodeName = $tagName;
    $new->ownerDocument =& $this;
    return $new;
  }
  
  /**
   * Creates an EntityReference node (not supported yet)<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name <code>string</code>
   * @returns EntityReference
   */
  function &createEntityReference($name) {
    $new = new EntityReference();
    $new->nodeName = $name;
    $new->ownerDocument =& $this;
    return $new;
  }
  
  /**
   * Creates a ProcessingInstruction node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $target <code>string</code>
   * @param   $data   <code>string</code>
   * @returns ProcessingInstruction
   */
  function &createProcessingInstruction($target, $data) {
    $new = new ProcessingInstruction();
    $new->nodeValue = $data;
    $new->nodeName = $target;
    $new->ownerDocument =& $this;
    return $new;
  }
        
  /**
   * Creates a Text node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $data <code>string</code>
   * @returns Text
   */
  function &createTextNode($data) {
    $new = new Text();
    $new->nodeValue = $data;
    $new->ownerDocument =& $this;
    return $new;
  }

  /**
   * Returns a NodeList of Element nodes from the current document whose 
   * tagName attribute matches the given tagname.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $tagName <code>string</code>
   * @returns NodeList
   */
  function &getElementsByTagName($tagName) {
    $result = new NodeList();
    $this->getElementsByTagNameList($tagName, $this->documentElement, $result);
    return $result;
  }
  
  /**
   * Same as createAttribute, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI  <code>string</code>
   * @param   $qualifiedName <code>string</code>
   * @returns Attr
   */
  function &createAttributeNS($namespaceURI, $qualifiedName) {
    $prefix = substr($qualifiedName, 0, strpos($qualifiedName, ":"));
    $localName = substr($qualifiedName, strpos($qualifiedName, ":")+1);
    
    if (empty($prefix) || empty($localName)) {
      return NAMESPACE_ERR;
    }

    if (empty($namespaceURI)) {
      return NAMESPACE_ERR;
    }

    if (($namespaceURI != "http://www.w3.org/XML/1998/namespace") && 
        ($prefix == "xml")) {
      return NAMESPACE_ERR;
    }

    if (($this->namespaceURI != "http://www.w3.org/2000/xmlns/") && 
        ($prefix == "xmlns")) {
      return NAMESPACE_ERR;
    }

    if ($this->qualifiedName == "xmlns") {
      return NAMESPACE_ERR;
    }

    $newAttr =& $this->createAttribute($qualifiedName);
    $newAttr->namespaceURI = $namespaceURI;
    $newAttr->localName = $localName;
    $newAttr->prefix = $prefix;
    return $newAttr;
  }

  /**
   * Same as createElement, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI  <code>string</code>
   * @param   $qualifiedName <code>string</code>
   * @returns Element
   */
  function &createElementNS($namespaceURI, $qualifiedName) {
    $prefix = substr($qualifiedName, 0, strpos($qualifiedName, ":"));
    $localName = substr($qualifiedName, strpos($qualifiedName, ":")+1);

    if (empty($prefix) || empty($localName)) {
      return NAMESPACE_ERR;
    }

    if (empty($namespaceURI)) {
      return NAMESPACE_ERR;
    }

    if (($namespaceURI != "http://www.w3.org/XML/1998/namespace") && 
        ($prefix == "xml")) {
      return NAMESPACE_ERR;
    }

    $newElem =& $this->createElement($qualifiedName);
    $newElem->namespaceURI = $namespaceURI;
    $newElem->localName = $localName;
    $newElem->prefix = $prefix;
    return $newElem;
  }

  /**
   * Returns a Element node from the current document whose id attribute 
   * matches the given id.<br>
   * DOM-Level 2 -- NYI
   *
   * @public
   * @param   $elementID <code>string</code>
   * @returns Element
   */
  function &getElementById($elementID) {
    $element = null;
    if (isset($this->doctype) &&
	isset($this->doctype->elements)) {
      $element = $this->getElementByIdHelp($elementID, true, 
					   $this->documentElement); 
    }
    else {
      $element = $this->getElementByIdHelp($elementID, false, 
					   $this->documentElement);
    }
    if ($element == null) {
      return NOT_FOUND_ERR;
    }
    else {
      return $element;
    }
  }

  /**
   * Same as getElementByTagName, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns NodeList
   */
  function &getElementsByTagNameNS($namespaceURI, $localName) {
    $result = new NodeList();
    $this->getElementsByTagNameNSList($namespaceURI, $localName, 
                                      $this->documentElement, $result);
    return $result;
  }
  
  /**
   * Creates a copy of a Node object from another document that can be
   * inserted within the current documents node hierarchy.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $importedNode <a href="Node.html">Node</a>
   * @param   $deep         <code>boolean</code>
   * @returns Node
   */
  function &importNode(&$importedNode, $deep) {
    if ($importedNode->nodeType == ATTRIBUTE_NODE) {
      $newNode =& $importedNode->cloneNode(true, $this);
      $newNode->ownerElement = NULL;
      $newNode->specified = true;
    }
    else {
      if (($importedNode->nodeType == ELEMENT_NODE) ||
	  ($importedNode->nodeType == ENTITY_NODE) ||
          ($importedNode->nodeType == DOCUMENT_FRAGMENT_NODE) || 
          ($importedNode->nodeType == PROCESSING_INSTRUCTION_NODE) ||
          ($importedNode->nodeType == TEXT_NODE) ||
          ($importedNode->nodeType == CDATA_SECTION_NODE) ||
          ($importedNode->nodeType == COMMENT_NODE)) {
        $newNode =& $importedNode->cloneNode($deep, $this);
      }
      else {
        return NOT_SUPPORTED_ERR;
      }
    }

    $newNode->ownerDocument =& $this;
    return $newNode;
  }

  /**
   * Runs recursively through the DOM-Tree and returns the element with 
   * attribute id.
   *
   * @private
   * @param   $tagName <code>string</code>
   * @param   $id      <code>boolean</code>
   * @param   $parent  <a href="Node.html">Node</a>
   * @returns void
   */
  function &getElementByIdHelp($id, $dtd, &$parent) {
    $element = null;
    if ($parent->nodeType == ELEMENT_NODE) {
      if (!$dtd) {
	$attr = $parent->getAttribute("id");
	if ($attr == $id) {
	  return $parent;
	}
      }
      else {
	$dtdElement = 
	   $this->doctype->elements->getNamedItem($parent->nodeName);
	if (isset($dtdElement) &&
	    isset($dtdElement->attributes)) {
	  for ($i = 0; $i < $dtdElement->attributes->length; ++$i) {
	    $dtdAttribute = $dtdElement->attributes->item($i);
	    if ($dtdAttribute->type == "ID") {
	      $attr = $parent->getAttribute($dtdAttribute->nodeName);
	      if ($attr == $id) {
		return $parent;
	      }
	    }
	  }
	}
      }

      if ($parent->hasChildNodes()) {
        $element = $this->getElementByIdHelp($id, $dtd, 
					     $parent->firstChild);
      }
    }

    if ((isset($parent->nextSibling)) && ($element != null)) {
      return $this->getElementByIdHelp($id, $parent->nextSibling, $result);
    }
    return $element;
  }
  
  /**
   * Runs recursively through the DOM-Tree and returns a NodeList with all
   * Element nodes that have the searched tagname.
   *
   * @private
   * @param   $tagName <code>string</code>
   * @param   $parent  <a href="Node.html">Node</a>
   * @param   $result  <a href="NodeList.html">NodeList</a>
   * @returns void
   */
  function getElementsByTagNameList($tagName, &$parent, &$result) {
    if ($parent->nodeType == ELEMENT_NODE) {
      if (($parent->tagName == $tagName) || ($tagName == "*")) {
        $result->insertNode($parent);
      }

      if ($parent->hasChildNodes()) {
        $this->getElementsByTagNameList($tagName, $parent->firstChild, 
                                        $result);
      }
    }

    if (isset($parent->nextSibling)) {
      $this->getElementsByTagNameList($tagName, $parent->nextSibling, $result);
    }
  }
  
  /**
   * Runs recursively through the DOM-Tree and returns a NodeList with all
   * Element nodes that have the searched $namespaceURI and $localName.
   *
   * @private
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @param   $parent       <a href="Node.html">Node</a>
   * @param   $result       <a href="NodeList.html">NodeList</a>
   * @returns void
   */
  function getElementsByTagNameNSList($namespaceURI, $localName, &$parent, &$result) {
    if ($parent->nodeType == ELEMENT_NODE) {
      if ((($parent->namespaceURI == $namespaceURI) && 
           ($parent->localName == $localName)) ||
          (($namespaceURI == "*") && ($parent->localName == $localName)) ||
          (($parent->namespaceURI == $namespaceURI) && ($localName == "*")) ||
          (($namespaceURI == "*") && ($localName == "*"))) {
        $result->insertNode($parent);
      }

      if ($parent->hasChildNodes()) {
        $this->getElementsByTagNameNSList($namespaceURI, $localName, 
                                          $parent->firstChild, $result);
      }
    }
    
    if (isset($parent->nextSibling)) {
      $this->getElementsByTagNameNSList($namespaceURI, $localName, 
                                        $parent->nextSibling, $result);
    }
  }
}
?>
