<?php
// phpXMLDOM - an XML DOM Implementation
//
// This Software and all associated files are released unter the 
// GNU Public License (GPL), see LICENSE for details.
//
// $Id: Element.php,v 1.4 2002/02/13 21:21:47 thomi Exp $

/**
 * Class representing the DOM Element interface.
 *
 * @package  phpXD
 * @author   Thomas Dohmke <thomas@dohmke.de>
 * @version  $Revision: 1.4 $
 */
class Element extends Node {
  /**
   * The XML tag name.<br>
   * DOM-Level 1
   *
   * @private
   * @type    string
   */
  var $tagName;
  
  /**
   * Constructor of the class.
   *
   * @public
   */
  function Element() {
    $this->Node();
    $this->nodeType = ELEMENT_NODE;
    $this->tagName =& $this->nodeName;
  }
  
  /**
   * Returns the XML tag name of the element.<br>
   * DOM-Level 1
   *
   * @public
   * @returns string
   */
  function getTagName() {
    return $this->tagName;
  }
  
  /**
   * Returns the value of the attribute specified by the name parameter.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name <code>string</code>
   * @returns string
   */
  function getAttribute($name) {
    $attr =& $this->getAttributeNode($name);
    if ($attr != NOT_FOUND_ERR) {
      return $attr->getValue();
    }
    else {
      return "";
    }
  }
  
  /**
   * Returns the attribute specified by the name parameter.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name <code>string</code>
   * @returns Attr
   */
  function &getAttributeNode($name) {
    if (isset($this->attributes)) {
      return $this->attributes->getNamedItem($name);
    }
  }
  
  /**
   * Returns a NodeList of all child Element nodes whose tagName attribute
   * matches the give name parameter.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name <code>string</code>
   * @returns NodeList
   */
  function &getElementsByTagName($name) {
    $result = new NodeList();
    $this->getElementsByTagNameList($name, $this->firstChild, $result);
    return $result;
  }

  /**
   * Removes the Attribute node specified by name.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name <code>string</code>
   * @returns void
   */
  function removeAttribute($name) {
    if (isset($this->attributes)) {
      $attr =& $this->attributes->getNamedItem($name);
      $this->removeAttributeNode($attr);
    }
  }
  
  /**
   * Removes the specified Attribute node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $oldAttr <a href="Attr.html">Attr</a>
   * @returns Attr
   */
  function &removeAttributeNode(&$oldAttr) {
    $name = $oldAttr->nodeName;
    if (isset($oldAttr->namespaceURI)) {
      $namespaceURI = $oldAttr->namespaceURI;
    }
    else {
      $namespaceURI = "";
    }
    if (isset($this->attributes)) {
      if ($namespaceURI == "") {
	$oldAttr =& $this->attributes->removeNamedItem($oldAttr->nodeName);
      }
      else {
	$localName = substr($name, strpos($name, ":") + 1);
	$oldAttr =& $this->attributes->removeNamedItemNS($namespaceURI, 
							 $localName);
      }
    }
    if (isset($this->ownerDocument->doctype)) {
      $dtdElement =& 
	 $this->ownerDocument->doctype->elements->
	 getNamedItem($this->nodeName);
      $dtdAttribute =& $dtdElement->attributes->getNamedItem($name);
      if (($dtdAttribute->default == "#FIXED") ||
	  ($dtdAttribute->defaultValue != "")) {
	if (empty($namespaceURI)) {
	  $this->setAttribute($name, $dtdAttribute->defaultValue);
	  $attribute =& $this->getAttributeNode($name);
	  $attribute->specified = false;
	} 
	else {
	  $this->setAttributeNS($namespaceURI, $name, 
				$dtdAttribute->defaultValue);
	  $attribute =& $this->getAttributeNodeNS($namespaceURI, $localName);
	  $attribute->specified = false;
	}
      }
    }
    return $oldAttr;
  }
  
  /**
   * Set the Attribute name to the value. If the node does not exist, it is
   * created.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $name  string
   * @param   $value string
   * @returns void
   */
  function setAttribute($name, $value) {
    $new = new Attr();
    $new->nodeName = $name;
    $new->setValue($value);
    $new->ownerElement = &$this;
    $new->ownerDocument = &$this->ownerDocument;
    $new->specified = true;
    
    if (empty($this->attributes)) {
      $this->attributes = new NamedNodeMap();
    }
    
    $this->attributes->setNamedItem($new);
  }

  /**
   * Sets the specified Attribute node.<br>
   * DOM-Level 1
   *
   * @public
   * @param   $newAttr <a href="Attr.html">Attr</a>
   * @returns Attr
   */
  function &setAttributeNode(&$newAttr) {
    $newAttr->specified = true;

    if (empty($this->attributes)) {
      $this->attributes = new NamedNodeMap();
    }
    
    return $this->attributes->setNamedItem($newAttr);
  }

  /**
   * Same as getAttribute, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns string
   */
  function getAttributeNS($namespaceURI, $localName) {
    if (isset($this->attributes)) {
      $attr =& $this->attributes->getNamedItemNS($namespaceURI, $localName);
      return $attr->getValue();
    }
  }
  
  /**
   * Same as getAttributeNode, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns Attr
   */
  function &getAttributeNodeNS($namespaceURI, $localName) {
    if (isset($this->attributes)) {
      return $this->attributes->getNamedItemNS($namespaceURI, $localName);
    }
  }
  
  /**
   * Same as getElementsByTagName, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns NodeList
   */
  function &getElementsByTagNameNS($namespaceURI, $localName) {
    $result = new NodeList();
    $this->getElementsByTagNameNSList($namespaceURI, $localName, 
				      $this->firstChild, $result);
    return $result;
  }

  /**
   * Returns true when an attribute with a given name is specified on this 
   * element or has a default value, false otherwise. <br>
   * DOM-Level 2
   *
   * @public
   * @param   $name   <code>string</code>
   * @returns boolean
   */
  function hasAttribute($name) {
    if (empty($this->attributes)) {
      return false;
    }

    $attr =& $this->attributes->getNamedItem($name);
    if (!empty($attr)) {
      return true;
    }
    
    return false;
  }
  
  /**
   * Same as hasAttribute, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns boolean
   */
  function hasAttributeNS($namespaceURI, $localName) {
    if (empty($this->attributes)) {
      return false;
    }
    
    $attr =& $this->attributes->getNamedItemNS($namespaceURI, $localName);
    if (!empty($attr)) {
      return true;
    }
    
    return false;
  }

  /**
   * Same as removeAttribute, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns void
   */
  function removeAttributeNS($namespaceURI, $localName) {
    $attr =& $this->getNamedItemNS($namespaceURI, $localName);
    return $this->removeAttributeNode($attr);
  }
  
  /**
   * Same as setAttribute, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @returns void
   */
  function setAttributeNS($namespaceURI, $qualifiedName, $value) {
    $prefix = substr($qualifiedName, 0, strpos($qualifiedName, ":"));
    $localName = substr($qualifiedName, strpos($qualifiedName, ":") + 1);
    
    if (empty($prefix) || empty($localName)) {
      return NAMESPACE_ERR;
    }

    if (empty($namespaceURI)) {
      return NAMESPACE_ERR;
    }

    if (($namespaceURI != "http://www.w3.org/XML/1998/namespace") && 
        ($prefix == "xml")) {
      return NAMESPACE_ERR;
    }

    if (($namespaceURI != "http://www.w3.org/2000/xmlns/") && 
        ($prefix == "xmlns")) {
      return NAMESPACE_ERR;
    }

    if ($qualifiedName == "xmlns") {
      return NAMESPACE_ERR;
    }

    $new = new Attr();
    $new->namespaceURI = $namespaceURI;
    $new->nodeName = $qualifiedName;
    $new->prefix = $prefix;
    $new->localName = $localName;
    $new->setValue($value);
    $new->ownerElement = &$this;
    $new->ownerDocument = &$this->ownerDocument;
    $new->specified = true;
    
    if (empty($this->attributes)) {
      $this->attributes = new NamedNodeMap();
    }
    
    $this->attributes->setNamedItemNS($new);
  }

  /**
   * Same as setAttributeNode, but includes support for XML namespaces.<br>
   * DOM-Level 2
   *
   * @public
   * @param   $newAttr <a href="Attr.html">Attr</a>
   * @returns Attr
   */
  function setAttributeNodeNS(&$newAttr) {
    $newAttr->specified = true;

    if (empty($this->attributes)) {
      $this->attributes = new NamedNodeMap();
    }
    
    return $this->attributes->setNamedItemNS($newAttr);
  }
  
  /**
   * Runs recursively through the DOM-Tree and returns a NodeList with all
   * Element nodes that have the searched tagname.
   *
   * @private
   * @param   $tagName <code>string</code>
   * @param   $parent  <a href="Node.html">Node</a>
   * @param   $result  <a href="NodeList.html">NodeList</a>
   * @returns void
   */
  function getElementsByTagNameList($tagName, &$parent, &$result) {
    if ($parent->nodeType == ELEMENT_NODE) {
      if (($parent->tagName == $tagName) || ($tagName == "*")) {
        $result->insertNode($parent);
      }
      
      if ($parent->hasChildNodes()) {
        $this->getElementsByTagNameList($tagName, $parent->firstChild, 
                                        $result);
      }
    }

    if (isset($parent->nextSibling)) {
      $this->getElementsByTagNameList($tagName, $parent->nextSibling, $result);
    }
  }

  /**
   * Runs recursively through the DOM-Tree and returns a NodeList with all
   * Element nodes that have the searched $namespaceURI and $localName.
   *
   * @private
   * @param   $namespaceURI <code>string</code>
   * @param   $localName    <code>string</code>
   * @param   $parent       <a href="Node.html">Node</a>
   * @param   $result       <a href="NodeList.html">NodeList</a>
   * @returns void
   */
  function getElementsByTagNameNSList($namespaceURI, $localName, &$parent, &$result) {
    if ($parent->nodeType == ELEMENT_NODE) {
      if ((($parent->namespaceURI == $namespaceURI) && 
           ($parent->localName == $localName)) ||
          (($namespaceURI == "*") && ($parent->localName == $localName)) ||
          (($parent->namespaceURI == $namespaceURI) && ($localName == "*")) ||
          (($namespaceURI == "*") && ($localName == "*"))) {
        $result->insertNode($parent);
      }
      
      if ($parent->hasChildNodes()) {
        $this->getElementsByTagNameNSList($namespaceURI, $localName, 
                                          $parent->firstChild, $result);
      }
    }
    
    if (isset($parent->nextSibling)) {
      $this->getElementsByTagNameNSList($namespaceURI, $localName, 
                                        $parent->nextSibling, $result);
    }
  }
}
?>
