<?php
// phpXMLDOM - an XML DOM Implementation
//
// This Software and all associated files are released unter the 
// GNU Public License (GPL), see LICENSE for details.
//
// $Id: DOMParser.php,v 1.7 2002/02/13 21:21:48 thomi Exp $

/**
 * Parse a document.
 *
 * @package phpXD
 * @author  Thomas Dohmke <thomas@dohmke.de>
 * @version $Revision: 1.7 $
 */
class DOMParser {
  /**
   * The Document node.
   *
   * @private
   * @type    Document
   */
  var $document;
  
  /**
   * The current Element node which is parsed. Only valid while
   * parsing an XML file!
   *
   * @private
   * @type    Node
   */
  var $currentNode;

  /**
   * The last node which is parsed. Only valid while parsing an XML file!
   *
   * @private
   * @type    Node
   */
  var $lastNode;
  
  /**
   * A CDataSection is parsed. All strings will be collected to one 
   * CDataSection node. Only valid while parsing an XML file!
   *
   * @private
   * @type    boolean
   */
  var $parseCData = false;
  
  /**
   * A array with the defined namespaces.
   *
   * @private
   * @type    array
   */
  var $namespaces = false;

  /**
   * If true, the DTD will be parsed.
   *
   * @private
   * @type    boolean
   */
  var $parseDTD = false;

  /**
   * Flag for handleDefault, that tokens are inside the dtd.
   *
   * @private
   * @type    boolean
   */
  var $DTDDecl = false;

  /**
   * Flag for handleDefault, that a internal dtd is parsed.
   *
   * @private
   * @type    boolean
   */
  var $DTDInternal = false;

  /**
   * Reference to a DTDParser.
   *
   * @private
   * @type    DTDParser
   */
  var $DTDParser = null;

  /**
   * An array with all tokens from dtd.
   *
   * @private
   * @type    array
   */
  var $DTDTokens;

  /**
   * Same as $DTDTokens but as flat string.
   *
   * @private
   * @type    string
   */
  var $DTDString = "";
  
  /**
   * The Constructor of the class. 
   *
   * @public
   * @returns void
   */
  function DOMParser() {
    $this->namespaces["xml"] = "http://www.w3.org/XML/1998/namespace";
    $this->namespaces["xmlns"] = "http://www.w3.org/2000/xmlns/";
  }
  
  /**
   * Loads the XML file specified by $file and creates the DOM tree.
   *
   * @public
   * @param   $file <code>string</code>
   * @param   $parseDTD <code>boolean</code>
   * @returns Document
   */
  function &parseFile($file, $parseDTD = false) {
    if ($this->file_exists($file)) {
      $content = implode("", file($file));
      
      if (!empty($content)) {
	return $this->parse($content, $parseDTD);
      }
    }
    else {
      $this->displayError("File %s could not be found or read.", $file);
    }
  }

  /**
   * Loads XML string specified by $str and creates the DOM tree.
   *
   * @public
   * @param   $str <code>string</code>
   * @param   $parseDTD <code>boolean</code>
   * @returns Document
   */
  function &parse($str, $parseDTD = false) {
    if (!empty($str)) {
      // Just a hack, cause it seems that expat doesn't support IGNORE or
      // INCLUDE.
      $str = preg_replace('=<!\[IGNORE\[.*\]\]>=sU', '', $str);
      $str = preg_replace('=<!\[INCLUDE\[(.*)\]\]>=sU', '\\1', $str);

      $this->document = new Document();
      if ($parseDTD) {
	$this->DTDParser = new DTDParser();
	$this->parseDTD = true;
      }
        
      $parser = xml_parser_create();
      
      xml_parser_set_option($parser, XML_OPTION_CASE_FOLDING, 0);
        
      xml_set_object($parser, $this);
        
      xml_set_element_handler($parser, "handleStartElement", 
			      "handleEndElement");
      xml_set_character_data_handler($parser, "handleCharacterData");
      xml_set_processing_instruction_handler($parser, 
					     "handleProcessingInstruction");
      xml_set_default_handler($parser, "handleDefault");
      
      if (!xml_parse($parser, $str)) {
	$this->displayError("XML error in line %d: %s", 
			    xml_get_current_line_number($parser), 
			    xml_error_string(xml_get_error_code($parser)));
      }
      
      xml_parser_free($parser);
      $this->document->normalize();
      return $this->document;
    }
    else {
      $this->displayError("The string is empty.");
    }
  }
  
  /**
   * Internal method to load a file of a parsed external entity.
   *
   * @private
   * @param   $file <code>string</code>
   * @returns Document
   */
  function &parseExternalEntity($file) {
    if ($this->file_exists($file)) {
      $content = implode("", file($file));
      
      if (!empty($content)) {
	$content = preg_replace('=<\?xml.*\?>\n=sU', '', $content);
	$content = preg_replace('=<\?xml.*\?>=sU', '', $content);
	return $this->parseEntity($content);
      }
    }
    else {
      $this->displayError("File %s could not be found or read.", $file);
    }
  }

  /**
   * Internal method to parse an entity body.
   *
   * @private
   * @param   $str <code>string</code>
   * @returns Document
   */
  function &parseEntity($str) {
    return $this->parse("<entity>".$str."</entity>");
  }

  /**
   * The Element Start Handler for Expat.
   *
   * @private
   * @returns void
   */
  function handleStartElement($parser, $name, $attributes) {
    if (!(strpos($name, ":") === false)) {
      $prefix = substr($name, 0, strpos($name, ":"));
      if (is_array($attributes) && 
          !empty($attributes["xmlns:".$prefix])) {
        $namespace = $attributes["xmlns:".$prefix];
      }
      else {
        if (is_array($this->namespaces) && 
            !empty($this->namespaces[$prefix])) {
          $namespace = $this->namespaces[$prefix];
        }
        else {
          $this->displayError("XML error: unknown namespace in line %d.", 
                              xml_get_current_line_number($parser));
        }
      }
      
      $this->namespaces[$prefix] = $namespace;
    }

    if (!(strpos($name, ":") === false)) {
      $newChild =& $this->document->createElementNS($namespace, $name);
      
      if ($newChild == NAMESPACE_ERR) {
	$this->displayError("XML error: namespace not valid in line %d.", 
			    xml_get_current_line_number($parser));
      }
    }
    else {
      $newChild =& $this->document->createElement($name);
    }

    if (!isset($this->document->documentElement)) {
      $this->document->documentElement =& $newChild;
      $this->document->appendChild($newChild);
      $this->currentNode =& $this->document->documentElement;
    }
    else {
      $this->currentNode->appendChild($newChild);
      $this->currentNode =& $newChild;
    }

    if (is_array($attributes)) {
      while (list($name, $value) = each($attributes)) {
        if (!(strpos($name, ":") === false)) {
          $prefix = substr($name, 0, strpos($name, ":"));
          $localName = substr($name, strpos($name, ":")+1);
          if ($prefix == "xmlns") {
            $this->namespaces[$localName] = $value;
            $namespace = $this->namespaces["xmlns"];
          }
          else {
            if (is_array($attributes) && 
                !empty($attributes["xmlns:".$prefix])) {
              $namespace = $attributes["xmlns:".$prefix];
            }
            else {
              if (is_array($this->namespaces) && 
                  !empty($this->namespaces[$prefix])) {
                $namespace = $this->namespaces[$prefix];
              }
              else {
                $this->displayError("XML error: unknown namespace".
                                    " in line %d.", 
                                    xml_get_current_line_number($parser));
              }
            }
          }

          $result = $this->currentNode->setAttributeNS($namespace, $name, 
                                                       $value);
	  $attribute =& $this->currentNode->getAttributeNodeNS($namespace,
							       $localName);

          if ($result == NAMESPACE_ERR) {
            $this->displayError("XML error: namespace not valid in line %d.", 
                                xml_get_current_line_number($parser));
          }
        }
        else {
          $this->currentNode->setAttribute($name, $value);
	  $attribute =& $this->currentNode->getAttributeNode($name);
        }

	if (isset($this->document->doctype)) {
	  $dtdElement =& 
	     $this->document->doctype->elements->
	     getNamedItem($this->currentNode->nodeName);
	  if (isset($dtdElement->attributes)) {
	    $dtdAttribute =& $dtdElement->attributes->getNamedItem($name);
	    if ($dtdAttribute->default == "#FIXED") {
	      // Because the xml parser do not differ between fixed and
	      // specified attributes, all fixed attributes are not
	      // specified, also if there are specified.
	      $attribute->specified = false;
	    }
	    else {
	      if ((isset($dtdAttribute->defaultValue)) &&
		  ($attribute->value == $dtdAttribute->defaultValue)) {
		$attribute->specified = false;
	      }
	    }
	  }
	}
      }
    }

    $this->lastNode =& $this->currentNode;
  }

  /**
   * The Element End Handler for Expat.
   *
   * @private
   * @returns void
   */
  function handleEndElement($parser, $name) {
    $this->currentNode =& $this->currentNode->parentNode;
    $this->lastNode =& $this->currentNode;
  }

  /**
   * The CharacterData Handler for Expat.
   *
   * @private
   * @returns void
   */
  function handleCharacterData($parser, $text) {
    if ($this->parseCData) {
      if ($this->lastNode->nodeType == CDATA_SECTION_NODE) {
        $this->lastNode->appendData($text);
      }
      else {
        $this->lastNode =& $this->currentNode->appendChild(
                              $this->document->createCDataSection($text));
      }
    }
    else {
      $this->lastNode =& $this->currentNode->appendChild(
	         	    $this->document->createTextNode($text));
    }
  }

  /**
   * The ProcessingInstruction Handler for Expat.
   *
   * @private
   * @returns void
   */
  function handleProcessingInstruction($parser, $target, $data) {
    if (!isset($this->document->documentElement)) {
      $this->document->appendChild(
         $this->document->createProcessingInstruction($target, $data));
    }
    else {
      $this->lastNode =& $this->currentNode->appendChild(
                            $this->document->createProcessingInstruction(
                               $target, $data));
    }
  }

  /**
   * The EntityReference Handler for Expat.
   *
   * @private
   * @returns void
   */
  function handleEntityReference($parser, $data) {
    $entityRef =& $this->document->createEntityReference($data);
    $this->lastNode =& $this->currentNode->appendChild($entityRef);
    $entity = $this->document->doctype->entities->getNamedItem($data);
    if (($entity != null) &&
	($entity->hasChildNodes())) {
      if ($entity != null) {
	for ($i = 0; $i < $entity->childNodes->length; ++$i) {
	  $entityRef->appendChild($entity->childNodes->item($i));
	}
      }
    }
  }

  /**
   * The Default Handler for Expat.
   *
   * @private
   * @returns void
   */
  function handleDefault($parser, $data) {
    $data2 = trim($data);
    
    if (preg_match('=&([a-z,A-Z,0-9]*);=sU', $data2, $ent)) {
      $this->handleEntityReference($parser, $ent[1]);
    }

    if ($data2 == "<!DOCTYPE") {
      $this->DTDTokens[] = $data2;
      $this->DTDString .= $data;
      $this->DTDDecl = true;
      return true;
    }

    if ($this->DTDDecl) {
      if ($data2 != "") {
	$this->DTDTokens[] = $data;
      }
      $this->DTDString .= $data;
      if ($data2 == "[") {
	$this->DTDInternal = true;
      }
      if ($data2 == "]") {
	$this->DTDInternal = false;
      }
      if (($data2 == ">") && (!$this->DTDInternal)) {
	$this->DTDDecl = false;

	$doctype = new DocumentType();
	$doctype->name = $this->DTDTokens[1];
	$doctype->publicId = "";
	$doctype->systemId = "";
	$doctype->internalSubset = "";
	$offset = 2;
	if ($this->DTDTokens[2] == "PUBLIC") {
	  $doctype->publicId = $this->DTDTokens[3];
	  $doctype->systemId = $this->DTDTokens[4];
	  $offset = 5;
	}
	if ($this->DTDTokens[2] == "SYSTEM") {
	  $doctype->systemId = $this->DTDTokens[3];
	  $offset = 4;
	}
	if ($this->DTDTokens[$offset] == "[") {
	  $subset = substr($this->DTDString, 
			   strpos($this->DTDString, "[") + 1);
	  $subset = substr($subset, 0, strrpos($subset, "]"));
	  $doctype->internalSubset = $subset;
	}
	$this->document->doctype =& $doctype;
	if ($this->parseDTD) {
	  $this->DTDParser->parseTokens($this->DTDTokens);
	  if ((isset($this->DTDParser->elements)) &&
	      ($this->DTDParser->elements->getLength() > 0)) {
	    $doctype->elements =& $this->DTDParser->elements;
	  }
	  if ((isset($this->DTDParser->entities)) &&
	      ($this->DTDParser->entities->getLength() > 0)) {
	    $doctype->entities = new NamedNodeMap();
	    for ($i = 0; $i < $this->DTDParser->entities->length; ++$i) {
	      $node =& $this->DTDParser->entities->item($i);
	      $doctype->entities->setNamedItem($this->document->
					       importNode($node,
							  true));
	    }
	  }
	  if ((isset($this->DTDParser->notations)) &&
	      ($this->DTDParser->notations->getLength() > 0)) {
	    $doctype->notations = new NamedNodeMap();
	    for ($i = 0; $i < $this->DTDParser->notations->length; ++$i) {
	      $node =& $this->DTDParser->notations->item($i);
	      $doctype->notations->setNamedItem($this->document->
						importNode($node,
							   true));
	    }
	  }
	}
      }
      return true;
    }

    if ($data2 == "<![CDATA[") {
      $this->parseCData = true;
      return true;
    }

    if ($data2 == "]]>" && $this->parseCData) {
      $this->parseCData = false;
      return true;
    }

    if (!(strpos($data2, "<!--") === false)) {
      $data2 = str_replace("<!--", "", $data2);
      $data2 = str_replace("-->", "", $data2);
      if (!isset($this->document->documentElement)) {
        $this->document->appendChild(
                  $this->document->createComment($data2));
      }
      else {
        $this->lastNode =& $this->currentNode->appendChild(
                              $this->document->createComment($data2));
      }
      return true;
    }

    return false;
  }

  /**
   * Check whether a file or url exists.
   *
   * @private
   * @returns void
   */
  function file_exists($filename) {
    $file = @fopen($filename, "r");
    if (!$file) {
      return false;
    }
    return true;
  }

  /**
   * Displays errors, which occur while parsing.
   *
   * @private
   * @returns void
   */
  function displayError($message) {
    if (func_num_args() > 1) {
      $arguments = func_get_args();
      
      $command = "\$message = sprintf(\$message, ";
      
      for ( $i = 1; $i < sizeof($arguments); $i++ ) {
        $command .= "\$arguments[".$i."], ";
      }

      $command = eregi_replace(", $", ");", $command);
      
      eval($command);
    }

    echo "<strong>DOMParser error:</strong> ".$message;
    exit;
  }
}
?>
